@@ -266,6 +266,7 @@ int acct_process(void)
 	unsigned long vsize, esp;
 
 	if (acct_active) {
+		memset(&ac, 0, sizeof(struct acct));
 		strncpy(ac.ac_comm, current->comm, ACCT_COMM);
 		ac.ac_comm[ACCT_COMM] = '\0';
 		ac.ac_utime = current->utime;
@@ -444,7 +445,7 @@ asmlinkage int sys_setreuid(uid_t ruid, uid_t euid)
 	if (ruid != (uid_t) -1 ||
 	    (euid != (uid_t) -1 && euid != old_ruid))
 		current->suid = current->euid;
-	current->fsuid = euid;
+	current->fsuid = current->euid;
 	return 0;
 }
 
